﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using System.Threading;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis.Symbolic;

// type definitions for ease of use
using CloneMapType = HeuristicLab.Core.ItemDictionary<HeuristicLab.Core.IItem, HeuristicLab.Core.IItem>;

namespace HeuristicLab.EvolutionaryTracking {
  /// <summary>
  /// Population diversity analyzer
  /// </summary>
  [Item("SymbolicExpressionTreePopulationDiversityAnalyzer", "An operator that tracks population diversity")]
  [StorableClass]
  public sealed class SymbolicExpressionTreePopulationDiversityAnalyzer : SingleSuccessorOperator, IAnalyzer {
    #region Parameter names
    private const string SymbolicExpressionTreeParameterName = "SymbolicExpressionTree";
    private const string UpdateIntervalParameterName = "UpdateInterval";
    private const string UpdateCounterParameterName = "UpdateCounter";
    private const string ResultsParameterName = "Results";
    private const string GenerationsParameterName = "Generations";
    private const string StoreHistoryParameterName = "StoreHistory";
    // comparer parameters
    private const string MatchVariablesParameterName = "MatchVariableNames";
    private const string MatchVariableWeightsParameterName = "MatchVariableWeights";
    private const string MatchConstantValuesParameterName = "MatchConstantValues";
    private const string SortSubtreesParameterName = "SortSubtrees";
    private const string SimilarityValuesParmeterName = "Similarity";
    // population graph
    private const string PopulationGraphParameterName = "PopulationGraph";
    // clone map
    private const string GlobalCloneMapParameterName = "GlobalCloneMap";

    #endregion

    #region Parameters

    public ValueParameter<IntValue> UpdateIntervalParameter {
      get { return (ValueParameter<IntValue>)Parameters[UpdateIntervalParameterName]; }
    }

    public ValueParameter<IntValue> UpdateCounterParameter {
      get { return (ValueParameter<IntValue>)Parameters[UpdateCounterParameterName]; }
    }

    public LookupParameter<ResultCollection> ResultsParameter {
      get { return (LookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }

    public LookupParameter<IntValue> GenerationsParameter {
      get { return (LookupParameter<IntValue>)Parameters[GenerationsParameterName]; }
    }

    public ValueParameter<BoolValue> StoreHistoryParameter {
      get { return (ValueParameter<BoolValue>)Parameters[StoreHistoryParameterName]; }
    }

    public IScopeTreeLookupParameter<ISymbolicExpressionTree> SymbolicExpressionTreeParameter {
      get { return (IScopeTreeLookupParameter<ISymbolicExpressionTree>)Parameters[SymbolicExpressionTreeParameterName]; }
    }

    public ValueParameter<BoolValue> MatchVariableNamesParameter {
      get { return (ValueParameter<BoolValue>)Parameters[MatchVariablesParameterName]; }
    }

    public ValueParameter<BoolValue> MatchVariableWeightsParameter {
      get { return (ValueParameter<BoolValue>)Parameters[MatchVariableWeightsParameterName]; }
    }

    public ValueParameter<BoolValue> MatchConstantValuesParameter {
      get { return (ValueParameter<BoolValue>)Parameters[MatchConstantValuesParameterName]; }
    }

    public ValueParameter<BoolValue> SortSubtreesParameter {
      get { return (ValueParameter<BoolValue>)Parameters[SortSubtreesParameterName]; }
    }

    public LookupParameter<CloneMapType> GlobalCloneMapParameter {
      get { return (LookupParameter<CloneMapType>)Parameters[GlobalCloneMapParameterName]; }
    }

    public ILookupParameter<DoubleValue> SimilarityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters[SimilarityValuesParmeterName]; }
    }

    #endregion

    #region Parameter properties

    public IntValue UpdateInterval {
      get { return UpdateIntervalParameter.Value; }
    }

    public IntValue UpdateCounter {
      get { return UpdateCounterParameter.Value; }
    }

    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }

    public CloneMapType GlobalCloneMap {
      get { return GlobalCloneMapParameter.ActualValue; }
    }

    public IntValue Generations {
      get { return GenerationsParameter.ActualValue; }
    }

    public BoolValue StoreHistory {
      get { return StoreHistoryParameter.Value; }
    }

    #endregion

    [StorableConstructor]
    private SymbolicExpressionTreePopulationDiversityAnalyzer(bool deserializing)
      : base(deserializing) {
    }

    private SymbolicExpressionTreePopulationDiversityAnalyzer(
      SymbolicExpressionTreePopulationDiversityAnalyzer original, Cloner cloner)
      : base(original, cloner) {
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SymbolicExpressionTreePopulationDiversityAnalyzer(this, cloner);
    }

    public SymbolicExpressionTreePopulationDiversityAnalyzer() {
      // add parameters
      Parameters.Add(new ScopeTreeLookupParameter<ISymbolicExpressionTree>(SymbolicExpressionTreeParameterName, "The symbolic expression trees to analyze."));
      Parameters.Add(new ValueParameter<IntValue>(UpdateIntervalParameterName, "The interval in which the tree length analysis should be applied.", new IntValue(1)));
      Parameters.Add(new ValueParameter<IntValue>(UpdateCounterParameterName, "The value which counts how many times the operator was called since the last update", new IntValue(0)));
      Parameters.Add(new ValueLookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
      Parameters.Add(new LookupParameter<IntValue>(GenerationsParameterName, "The number of generations so far."));
      Parameters.Add(new ValueParameter<BoolValue>(MatchVariablesParameterName, "Specify if the symbolic expression tree comparer should match variable names."));
      Parameters.Add(new ValueParameter<BoolValue>(MatchVariableWeightsParameterName, "Specify if the symbolic expression tree comparer should match variable weights."));
      Parameters.Add(new ValueParameter<BoolValue>(MatchConstantValuesParameterName, "Specify if the symbolic expression tree comparer should match constant values."));
      Parameters.Add(new ValueParameter<BoolValue>(SortSubtreesParameterName, "Specifies whether the subtrees of a tree should be sorted before comparison."));
      Parameters.Add(new LookupParameter<CloneMapType>(GlobalCloneMapParameterName, "A global map keeping track of trees and their clones (made during selection)."));
      Parameters.Add(new ValueParameter<BoolValue>(StoreHistoryParameterName, "True if the tree lengths history of the population should be stored.", new BoolValue(false)));
      Parameters.Add(new LookupParameter<DoubleValue>(SimilarityValuesParmeterName, ""));
      UpdateCounterParameter.Hidden = true;
      UpdateIntervalParameter.Hidden = true;
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
    }

    #region IStatefulItem members
    public override void InitializeState() {
      base.InitializeState();
      UpdateCounter.Value = 0;
    }

    public override void ClearState() {
      base.ClearState();
      UpdateCounter.Value = 0;
    }
    #endregion

    public bool EnabledByDefault {
      get { return true; }
    }

    public override IOperation Apply() {
      if (SimilarityParameter.ActualValue == null || SimilarityParameter.ActualValue.Value.IsAlmost(-1.0)) {
        UpdateCounter.Value++;
        if (UpdateCounter.Value != UpdateInterval.Value) return base.Apply();

        UpdateCounter.Value = 0;
        var trees = SymbolicExpressionTreeParameter.ActualValue.ToList();
        if (SortSubtreesParameter.Value.Value) {
          var canonicalSorter = new SymbolicExpressionTreeCanonicalSorter();
          foreach (var t in trees) canonicalSorter.SortSubtrees(t);
        }

        SimilarityParameter.ActualValue = new DoubleValue();

        var comp = new SymbolicExpressionTreeNodeSimilarityComparer {
          MatchConstantValues = MatchConstantValuesParameter.Value.Value,
          MatchVariableNames = MatchVariableNamesParameter.Value.Value,
          MatchVariableWeights = MatchVariableWeightsParameter.Value.Value
        };

        var operations = new OperationCollection { Parallel = true };
        foreach (var tree in trees) {
          var op = new SymbolicDataAnalysisExpressionTreeSimilarityCalculator { CurrentSymbolicExpressionTree = tree, SimilarityComparer = comp };
          var operation = ExecutionContext.CreateChildOperation(op, ExecutionContext.Scope);
          operations.Add(operation);
        }
        return new OperationCollection { operations, ExecutionContext.CreateOperation(this) };
      }

      ResultCollection results = ResultsParameter.ActualValue;
      // population diversity
      DataTable populationDiversityTable;
      if (!results.ContainsKey("PopulationDiversity")) {
        populationDiversityTable = new DataTable("PopulationDiversity") { VisualProperties = { YAxisTitle = "Diversity" } };
        results.Add(new Result("PopulationDiversity", populationDiversityTable));
      }
      populationDiversityTable = (DataTable)results["PopulationDiversity"].Value;
      if (!populationDiversityTable.Rows.ContainsKey("Diversity"))
        populationDiversityTable.Rows.Add(new DataRow("Diversity") { VisualProperties = { StartIndexZero = true } });

      int length = SymbolicExpressionTreeParameter.ActualValue.Length;
      var similarity = SimilarityParameter.ActualValue.Value / (length * (length - 1) / 2.0);
      var diversity = 1 - similarity;
      SimilarityParameter.ActualValue.Value = -1.0;

      populationDiversityTable.Rows["Diversity"].Values.Add(diversity);

      // selection diversity
      if (GlobalCloneMap == null) return base.Apply();
      DataTable relativeSelectionCountsTable;
      if (!results.ContainsKey("SelectedIndividuals")) {
        relativeSelectionCountsTable = new DataTable("SelectedIndividuals") { VisualProperties = { YAxisTitle = "% Selected Individuals" } };
        results.Add(new Result("SelectedIndividuals", relativeSelectionCountsTable));
      }
      relativeSelectionCountsTable = (DataTable)Results["SelectedIndividuals"].Value;
      if (!relativeSelectionCountsTable.Rows.ContainsKey("SelectedIndividuals")) {
        relativeSelectionCountsTable.Rows.Add(new DataRow("SelectedIndividuals") { VisualProperties = { StartIndexZero = true } });
      }
      double relativeSelectionCount = GlobalCloneMap.Values.Distinct().Count() / (double)SymbolicExpressionTreeParameter.ActualValue.Length;
      relativeSelectionCountsTable.Rows["SelectedIndividuals"].Values.Add(relativeSelectionCount);

      // build a selection frequencies histogram
      var graph = (SymbolicExpressionTreeGenealogyGraph)results[PopulationGraphParameterName].Value;
      // get graph nodes corresponding to individuals of the previous generation
      var prevGen = graph.Nodes.Where(node => node.Rank.IsAlmost(Generations.Value - 1)).OrderByDescending(n => n.Quality).ToList();
      var frequencies = new double[SymbolicExpressionTreeParameter.ActualValue.Length];
      for (int i = 0; i != prevGen.Count; ++i) {
        int selFreq = GlobalCloneMap.Values.Count(t => t == prevGen[i].SymbolicExpressionTree);
        frequencies[i] = selFreq;
      }
      DataTable selectionFrequenciesTable;
      if (!results.ContainsKey("SelectionFrequencies")) {
        selectionFrequenciesTable = new DataTable("Selection Frequencies") { VisualProperties = { YAxisTitle = "Selection Count" } };
        results.Add(new Result("SelectionFrequencies", selectionFrequenciesTable));
      }
      selectionFrequenciesTable = (DataTable)results["SelectionFrequencies"].Value;
      DataRow histoRow;
      if (!selectionFrequenciesTable.Rows.ContainsKey("SelectionFrequencies")) {
        histoRow = new DataRow("SelectionFrequencies") { VisualProperties = { StartIndexZero = true } };
        selectionFrequenciesTable.Rows.Add(histoRow);
      }
      histoRow = selectionFrequenciesTable.Rows["SelectionFrequencies"];
      histoRow.Values.Replace(frequencies);

      bool storeHistory = StoreHistory.Value;
      if (storeHistory) {
        DataTableHistory selectionFrequenciesHistory;
        if (!results.ContainsKey("SelectionFrequenciesHistory")) {
          selectionFrequenciesHistory = new DataTableHistory();
          results.Add(new Result("SelectionFrequenciesHistory", selectionFrequenciesHistory));
        }
        selectionFrequenciesHistory = (DataTableHistory)results["SelectionFrequenciesHistory"].Value;
        selectionFrequenciesHistory.Add((DataTable)selectionFrequenciesTable.Clone());
      }
      return base.Apply();
    }
  }
}
